# -*- coding: utf-8 -*-
"""
Decides whether the bot responds to a message.
"""

import random
import typing

from oobabot import fancy_logger
from oobabot import persona
from oobabot import types


class LastReplyTimes(dict):
    """
    A dictionary that keeps track of the last time we were mentioned
    in a channel.

    This uses the timestamp on the message, not the local system's
    RTC.  The advantage of this is that if messages are delayed,
    we'll only respond to ones that were actually sent within the
    appropriate time window.
    """

    def __init__(self, cache_timeout: float):
        self.cache_timeout = cache_timeout

    def purge_outdated(self, latest_timestamp: float) -> None:
        oldest_time_to_keep = latest_timestamp - self.cache_timeout
        purged = {
            channel_id: response_time
            for channel_id, response_time in self.items()
            if response_time >= oldest_time_to_keep
        }
        self.clear()
        self.update(purged)

    def log_mention(self, channel_id: int, send_timestamp: float) -> None:
        self[channel_id] = send_timestamp

    def time_since_last_mention(self, message: types.ChannelMessage) -> float:
        self.purge_outdated(message.send_timestamp)
        return message.send_timestamp - self.get(message.channel_id, 0)


class DecideToRespond:
    """
    Decide whether to respond to a message.
    """

    def __init__(
        self,
        discord_settings: dict,
        persona: persona.Persona,
        interrobang_bonus: float,
        time_vs_response_chance: typing.List[typing.Tuple[float, float]],
    ):
        self.ignore_dms = discord_settings["ignore_dms"]
        self.interrobang_bonus = interrobang_bonus
        self.persona = persona
        self.time_vs_response_chance = time_vs_response_chance

        last_reply_cache_timeout = max(time for time, _ in time_vs_response_chance)
        self.last_reply_times = LastReplyTimes(last_reply_cache_timeout)

    def is_directly_mentioned(
        self, our_user_id: int, message: types.GenericMessage
    ) -> bool:
        """
        Returns True if the message is a direct message to us, or if it
        mentions us by @name or wakeword.
        """

        # reply to all private messages
        if isinstance(message, types.DirectMessage):
            if self.ignore_dms:
                return False
            return True

        # reply to all messages in which we're @-mentioned
        if isinstance(message, types.ChannelMessage):
            if message.is_mentioned(our_user_id):
                return True

        # reply to all messages that include a wakeword
        if self.persona.contains_wakeword(message.body_text):
            return True

        return False

    def calc_base_chance_of_unsolicited_reply(
        self, message: types.ChannelMessage
    ) -> float:
        """
        Calculate base chance of unsolicited reply using the
        TIME_VS_RESPONSE_CHANCE table.
        """
        # return a base chance that we'll respond to a message that wasn't
        # addressed to us, based on the table in TIME_VS_RESPONSE_CHANCE.
        # other factors might increase this chance.
        time_since_last_send = self.last_reply_times.time_since_last_mention(message)
        response_chance = 0.0
        for duration, chance in self.time_vs_response_chance:
            if time_since_last_send < duration:
                response_chance = chance
                break
        return response_chance

    def provide_unsolicited_reply_in_channel(
        self, our_user_id: int, message: types.ChannelMessage
    ) -> bool:
        """
        Returns True if we should respond to the message, even
        though we weren't directly mentioned.
        """

        # if we're not at-mentioned but others are, don't reply
        if message.mentions and not message.is_mentioned(our_user_id):
            return False

        # if message is empty, don't reply.  This can happen if someone
        # posts an image or an attachment without a comment.
        if message.is_empty():
            return False

        # if we've posted recently in this channel, there are a few
        # other reasons we may respond.  But if we haven't, just
        # ignore the message.

        # if we haven't posted to this channel recently, don't reply
        response_chance = self.calc_base_chance_of_unsolicited_reply(message)
        if response_chance == 0.0:
            return False

        # if the new message ends with a question mark, we'll respond
        if message.body_text.endswith("?"):
            response_chance += self.interrobang_bonus

        # if the new message ends with an exclamation point, we'll respond
        if message.body_text.endswith("!"):
            response_chance += self.interrobang_bonus

        time_since_last_mention = self.last_reply_times.time_since_last_mention(message)
        fancy_logger.get().debug(
            "Considering unsolicited response in channel %s after %2.0f seconds.  "
            + "chance: %2.0f%%.",
            message.channel_name,
            time_since_last_mention,
            response_chance * 100.0,
        )

        if random.random() < response_chance:
            return True

        return False

    def should_reply_to_message(
        self, our_user_id: int, message: types.GenericMessage
    ) -> typing.Tuple[bool, bool]:
        """
        Returns a tuple of (should_reply, is_direct_mention).

        Direct mentions are always replied to, but also, the
        caller should log the mention later by calling log_mention().

        The only reason this method doesn't to so itself is that
        in the case of us generating a thread to reply on, the
        channel ID we want to track will be that of the thread
        we create, not the channel the message was posted in.
        """

        # ignore messages from other bots, out of fear of infinite loops,
        # as well as world domination
        if message.author_is_bot:
            return (False, False)

        # we do not want the bot to reply to itself.  This is redundant
        # with the previous check, except it won't be if someone decides
        # to run this under their own user token, rather than a proper
        # bot token.
        if message.author_id == our_user_id:
            return (False, False)

        if self.is_directly_mentioned(our_user_id, message):
            return (True, True)

        if isinstance(message, types.ChannelMessage):
            if self.provide_unsolicited_reply_in_channel(our_user_id, message):
                return (True, False)

        # ignore anything else
        return (False, False)

    def log_mention(self, channel_id: int, send_timestamp: float) -> None:
        self.last_reply_times.log_mention(channel_id, send_timestamp)



# -*- coding: utf-8 -*-
"""
Settings library which supports both CLI and YAML settings files.
Allows for settings to be exposed through one or both of these interfaces.

load - sets defaults, then loads from YAML, then loads from CLI.
load_from_cli - sets defaults, then loads from CLI only.
load_from_dict - sets defaults, then loads from a dictionary only.
load_from_yaml - sets defaults, then loads from YAML only.
write_to_file - writes the given config as YAML to the given file.
write_to_stream - writes the given config as YAML to the given stream.
"""

import argparse
import textwrap
import typing

import ruamel.yaml as ryaml

import oobabot

YAML_WIDTH = 88
DIVIDER = "# " * (YAML_WIDTH >> 1)
INDENT_UNIT = 2

SettingDictType = typing.Dict[
    str, typing.Union[bool, int, float, str, typing.List[str]]
]

SettingValueType = typing.Union[
    bool, int, float, str, typing.List[str], SettingDictType
]


def format_yaml_comment(comment_lines: typing.List[str]) -> str:
    out = []
    for line in comment_lines:
        out.append("\n".join(textwrap.wrap(line, width=YAML_WIDTH)))
    return "\n" + "\n".join(out)


def add_to_group(
    group: ryaml.CommentedMap,
    key: str,
    value: typing.Any,
    comment_lines: typing.List[str],
    indent: int,
) -> None:
    group[key] = value
    group.yaml_set_comment_before_after_key(
        key,
        before=format_yaml_comment(comment_lines),
        indent=indent,
    )


T = typing.TypeVar("T", bound="SettingValueType")


class ConfigSetting(typing.Generic[T]):
    """
    An individual setting that can be exposed through CLI and/or YAML
    """

    data_type: T
    description_lines: typing.List[str]
    cli_args: typing.List[str]
    include_in_argparse: bool
    include_in_yaml: bool
    show_default_in_yaml: bool
    fn_on_set: typing.Callable[[T], None]

    def __init__(
        self,
        name: str,
        default: T,
        description_lines: typing.List[str],
        cli_args: typing.Optional[typing.List[str]] = None,
        place_default_in_yaml: bool = False,
        include_in_argparse: bool = True,
        include_in_yaml: bool = True,
        show_default_in_yaml: bool = True,
        fn_on_set: typing.Callable[[T], None] = lambda x: None,
    ):
        self.name = name
        self.default = default
        self.description_lines = [x.strip() for x in description_lines]
        if cli_args is None:
            cli_args = ["--" + name.replace("_", "-")]
        self.cli_args = cli_args
        self.value = default
        self.include_in_argparse = include_in_argparse
        self.include_in_yaml = include_in_yaml
        self.place_default_in_yaml = place_default_in_yaml
        self.show_default_in_yaml = show_default_in_yaml
        self.fn_on_set = fn_on_set

    def add_to_argparse(self, parser: argparse._ArgumentGroup):
        if not self.include_in_argparse:
            return

        kwargs = {
            "default": self.value,
            "help": " ".join(self.description_lines),
        }

        # note: the other way to do this is with
        #   typing.get_args(self.__orig_class__)[0]
        # but that isn't officially supported, so
        # let's do it the jankier way
        if isinstance(self.default, str):
            kwargs["type"] = str
        elif isinstance(self.default, bool):
            kwargs["action"] = "store_true"
            if self.default:
                kwargs["action"] = "store_false"
        elif isinstance(self.default, int):
            kwargs["type"] = int
        elif isinstance(self.default, float):
            kwargs["type"] = float
        elif isinstance(self.default, list):
            kwargs["type"] = str
            kwargs["nargs"] = "*"

        parser.add_argument(*self.cli_args, **kwargs)

    def set_value_from_argparse(self, args: argparse.Namespace) -> None:
        if not self.include_in_argparse:
            return
        if not hasattr(args, self.name):
            raise ValueError(f"Namespace does not have attribute {self.name}")
        self.set_value(getattr(args, self.name))

    def add_to_yaml_group(self, group: ryaml.CommentedMap):
        if not self.include_in_yaml:
            return
        value = None
        if self.place_default_in_yaml or (self.value != self.default):
            value = self.value
        add_to_group(
            group,
            key=self.name,
            value=value,
            comment_lines=self.make_yaml_comment(),
            indent=INDENT_UNIT,
        )

    def make_yaml_comment(self) -> typing.List[str]:
        comment_lines = self.description_lines.copy()

        if self.show_default_in_yaml:
            if self.default is not None:
                comment_lines.append(f"  default: {self.default}")
            else:
                comment_lines.append("  default: None")
        return comment_lines

    def set_value_from_yaml(self, yaml: ryaml.CommentedMap) -> None:
        if not self.include_in_yaml:
            return
        if self.name not in yaml:
            return
        if (not self.place_default_in_yaml) and (yaml[self.name] is None):
            # if we didn't place the default in the yaml, and the setting
            # is now blank, that's no surprise.  Keep the default
            # rather than overwriting it with None
            return
        value = yaml[self.name]
        # if value is a dict, it's possible that it only contains
        # some, but not all of the keys that we need.  So we want to
        # merge it with the default dict, overwriting the default
        # values with the values from the yaml
        if isinstance(value, dict):
            if isinstance(self.default, dict):
                value = {**self.default, **value}
            else:
                raise ValueError(
                    f"Setting {self.name} is a dict, but the default value is not"
                )
        # note: keeping the type checker happy across python versions
        # here is actually kinda hard.  It's worried about what the
        # values in the combined dict might be.  But really it's fine,
        # so just tell it to chill.
        self.set_value(value)  # type: ignore

    def set_value(self, value: T) -> None:
        self.value = value
        self.fn_on_set(value)

    def get(self) -> T:
        if isinstance(self.value, dict):
            return self.value.copy()
        return self.value


class ConfigSettingGroup:
    """
    A group of related settings that can be exposed through CLI and/or YAML
    """

    name: str
    description: str
    settings: typing.Dict[str, ConfigSetting]

    def __init__(
        self,
        name: str,
        description: str = "",
        include_in_argparse: bool = True,
        include_in_yaml: bool = True,
    ):
        self.name = name
        self.description = description
        self.settings = {}
        self.include_in_argparse = include_in_argparse
        self.include_in_yaml = include_in_yaml

    def add_setting(self, setting: "ConfigSetting") -> None:
        self.settings[setting.name] = setting

    def add_to_argparse(self, parser: argparse.ArgumentParser):
        if not self.include_in_argparse:
            return
        arg_group = parser.add_argument_group(self.name, self.description)
        for setting in self.settings.values():
            setting.add_to_argparse(arg_group)

    def set_values_from_argparse(self, args: argparse.Namespace) -> None:
        if not self.include_in_argparse:
            return
        for setting in self.settings.values():
            setting.set_value_from_argparse(args)

    def add_to_yaml(self, yaml: ryaml.CommentedMap):
        if not self.include_in_yaml:
            return
        group_key = self.name.lower().replace(" ", "_")

        group = ryaml.CommentedMap()
        for setting in self.settings.values():
            setting.add_to_yaml_group(group)

        add_to_group(
            group=yaml,
            key=group_key,
            value=group,
            comment_lines=[DIVIDER, group_key, "."],
            indent=0,
        )

    def set_values_from_yaml(self, yaml: dict):
        if not self.include_in_yaml:
            return
        if yaml is None:
            return
        group_key = self.name.lower().replace(" ", "_")
        if group_key not in yaml:
            return
        group = yaml[group_key]
        for setting in self.settings.values():
            setting.set_value_from_yaml(group)

    def set_values_from_dict(self, main_dict: dict):
        group_dict = main_dict.get(self.name.lower().replace(" ", "_"), {})
        for setting in self.settings.values():
            if setting.name in group_dict:
                val = group_dict[setting.name]
                setting.set_value(val)

    def set(self, name: str, value: SettingValueType) -> None:
        self.settings[name].set_value(value)

    def get_setting(self, name: str) -> ConfigSetting:
        return self.settings[name]

    def get(self, name: str) -> SettingValueType:
        return self.settings[name].value

    def get_str(self, name: str) -> str:
        return self.settings[name].value

    def get_list(self, name: str) -> typing.List[SettingValueType]:
        return self.settings[name].value

    def get_all(self) -> typing.Dict[str, SettingValueType]:
        return {name: setting.get() for (name, setting) in self.settings.items()}.copy()


def load_from_yaml(
    filename: str,
    setting_groups: typing.List["ConfigSettingGroup"],
) -> None:
    """
    Load settings from a YAML file only
    """
    try:
        with open(filename, "r", encoding="utf-8") as file:
            yaml = ryaml.YAML(typ="safe")
            yaml_settings = yaml.load(file)
            for group in setting_groups:
                group.set_values_from_yaml(yaml_settings)

    except (FileNotFoundError, IsADirectoryError):
        pass


def load_from_cli(
    args,
    setting_groups: typing.List["ConfigSettingGroup"],
) -> argparse.ArgumentParser:
    """
    Load settings from the command line only
    """
    cli_parser = argparse.ArgumentParser(
        description=f"oobabot v{oobabot.__version__}: Discord bot for "
        + "oobabooga's text-generation-webui",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False,
    )

    # we need to initialize argparse AFTER we read from
    # the yaml file, so that the argparse defaults are
    # set to the yaml-read values.
    for group in setting_groups:
        group.add_to_argparse(cli_parser)

    cli_settings = cli_parser.parse_args(args=args)
    for group in setting_groups:
        group.set_values_from_argparse(cli_settings)

    return cli_parser


def load_from_dict(
    setting_groups: typing.List["ConfigSettingGroup"],
    settings_dict: dict,
) -> None:
    """
    Load settings from a dictionary
    """
    print(f"settings_dict: {settings_dict}")
    for group in setting_groups:
        group.set_values_from_dict(settings_dict)


def load(
    cli_args: typing.List[str],
    setting_groups: typing.List["ConfigSettingGroup"],
    config_file: str,
) -> argparse.ArgumentParser:
    """
    Load settings from defaults, config.yml, and command line arguments
    in that order.  Later sources will overwrite earlier ones.

    Returns the argparse parser, which can be used to print out the help
    message.
    """
    load_from_yaml(config_file, setting_groups)
    namespace = load_from_cli(cli_args, setting_groups)

    # returning this since it can print out the help message
    return namespace


START_COMMENT = textwrap.dedent(
    """
    # Welcome to Oobabot!
    #
    # This is the configuration file for Oobabot.  It is a YAML file, and
    # comments are allowed.  Oobabot attempts to load a file named
    # "config.yml" from the current directory when it is run.
    #
    """
)


def write_to_stream(
    setting_groups: typing.List["ConfigSettingGroup"],
    out_stream: typing.TextIO,
) -> None:
    yaml_map = ryaml.CommentedMap()
    yaml_map.yaml_set_start_comment(START_COMMENT)

    add_to_group(
        group=yaml_map,
        key="version",
        value=oobabot.__version__,
        comment_lines=[],
        indent=0,
    )

    for group in setting_groups:
        group.add_to_yaml(yaml_map)

    yaml = ryaml.YAML()
    yaml.default_flow_style = False
    yaml.allow_unicode = True
    yaml.indent(mapping=INDENT_UNIT, sequence=2 * INDENT_UNIT, offset=INDENT_UNIT)

    yaml.dump(yaml_map, out_stream)


def write_to_file(
    setting_groups: typing.List["ConfigSettingGroup"], filename: str
) -> None:
    """
    Write the current values in the setting groups to a YAML file
    """
    with open(filename, "w", encoding="utf-8") as file:
        write_to_stream(setting_groups, file)



# -*- coding: utf-8 -*-
"""
Generic types for messages, used to abstract away
Discord-specific types from the rest of the code.
"""
import typing


class GenericMessage:
    """
    Represents a message from a user.
    """

    def __init__(
        self,
        author_id: int,
        author_name: str,
        channel_id: int,
        channel_name: str,
        message_id: int,
        reference_message_id,
        body_text: str,
        author_is_bot: bool,
        send_timestamp: float,
    ):
        self.author_id = author_id
        self.author_name = author_name
        self.message_id = message_id
        self.body_text = body_text
        self.author_is_bot = author_is_bot
        self.reference_message_id = reference_message_id
        self.send_timestamp = send_timestamp
        self.channel_id = channel_id
        self.channel_name = channel_name

    def is_empty(self) -> bool:
        return not self.body_text.strip()


class DirectMessage(GenericMessage):
    """
    Represents a message sent directly to the bot.
    """


class ChannelMessage(GenericMessage):
    """
    Represents a message sent in a channel, including
    a private group chat or thread.
    """

    def __init__(
        self,
        /,
        mentions: typing.List[int],
        **kwargs,
    ):
        super().__init__(**kwargs)  # type: ignore
        self.mentions = mentions

    def is_mentioned(self, user_id: int) -> bool:
        return user_id in self.mentions



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""
Bot entrypoint.
"""

import asyncio
import concurrent.futures
import contextlib
import signal
import sys
import threading
import typing

import discord

from oobabot import bot_commands
from oobabot import decide_to_respond
from oobabot import discord_bot
from oobabot import discord_utils
from oobabot import fancy_logger
from oobabot import http_client
from oobabot import image_generator
from oobabot import ooba_client
from oobabot import persona
from oobabot import prompt_generator
from oobabot import repetition_tracker
from oobabot import response_stats
from oobabot import sd_client
from oobabot import settings
from oobabot import templates


# this warning causes more harm than good here
# pylint: disable=W0201
class Oobabot:
    """
    Main bot class.  Load settings, creates helper objects,
    and invokes the bot loop.

    Methods:
        constructor: Loads settings from the command line, environment
            variables, and config file.  This config may be changed
            before calling start().
        start: Start the bot.  Blocks until the bot exits.
        stop: Stop the bot.  Blocks until the bot exits.

    Class Methods:
        test_discord_token: Test a discord token to see if it's valid.
            Requires Internet access.
        generate_invite_url: Generate a URL that can be used to invite
            the bot to a server.
    """

    def __init__(
        self,
        cli_args: typing.List[str],
    ):
        """
        Initialize the bot, and load settings from the command line,
        environment variables, and config file.  These will be
        available in self.settings.

        self.settings is an :py:class:`oobabot.settings.Settings` object.
        """
        self.startup_lock = threading.Lock()
        self.settings = settings.Settings()

        self.settings.load(cli_args)

    def start(self):
        """
        Start the bot.  Blocks until the bot exits.

        When running from the CLI, the bot would normally exit
        when the user presses Ctrl-C.  Or otherwise sends a SIGINT
        or SIGTERM signal.

        This is also where one-off commands are run in CLI mode:
         - help: Print help and exit
         - generate_config: Print a the config file and exit
         - invite_url: Print a URL that can be used to invite the bot

        When running inside another process, the bot will exit
        when another thread calls stop().
        """

        with self.startup_lock:
            self._begin()

            if self.settings.general_settings.get("help"):
                self.settings.print_help()
                return

            if self.settings.general_settings.get("generate_config"):
                self.settings.write_to_stream(out_stream=sys.stdout)
                return

            if not self.settings.discord_settings.get("discord_token"):
                msg = (
                    f"Please set the '{self.settings.DISCORD_TOKEN_ENV_VAR}' "
                    + "environment variable to your bot's discord token."
                )
                print(msg, file=sys.stderr)
                if self._our_own_main():
                    sys.exit(1)
                else:
                    raise RuntimeError(msg)

            if self.settings.general_settings.get("invite_url"):
                url = self.generate_invite_url(
                    self.settings.discord_settings.get_str("discord_token")
                )
                print(url)
                return

            self._prepare_connections()

        asyncio.run(self._init_then_start())

    def _our_own_main(self) -> bool:
        """
        Returns True if we're running from the CLI, False if we're
        running inside another process.
        """
        return threading.current_thread() == threading.main_thread()

    def _begin(self):
        """
        Called after we've loaded our configuration but before we
        start the bot.  This is a good place to do any one-time setup
        for helper objects.
        """

        fancy_logger.init_logging(
            level=self.settings.discord_settings.get_str("log_level"),
            log_to_console=self._our_own_main(),
        )

        # templates used to generate prompts to send to the AI
        # as well as for some UI elements
        self.template_store = templates.TemplateStore(
            settings=self.settings.template_settings.get_all()
        )

        ########################################################
        # Connect to Oobabooga

        self.ooba_client = ooba_client.OobaClient(
            settings=self.settings.oobabooga_settings.get_all(),
        )

        ########################################################
        # Connect to Stable Diffusion, if configured

        self.stable_diffusion_client = None
        sd_settings = self.settings.stable_diffusion_settings.get_all()
        if sd_settings["stable_diffusion_url"]:
            self.stable_diffusion_client = sd_client.StableDiffusionClient(
                settings=sd_settings,
            )

        ########################################################
        # Bot logic

        self.persona = persona.Persona(
            persona_settings=self.settings.persona_settings.get_all()
        )

        # decides which messages the bot will respond to
        self.decide_to_respond = decide_to_respond.DecideToRespond(
            discord_settings=self.settings.discord_settings.get_all(),
            persona=self.persona,
            interrobang_bonus=self.settings.DECIDE_TO_RESPOND_INTERROBANG_BONUS,
            time_vs_response_chance=self.settings.TIME_VS_RESPONSE_CHANCE,
        )

        # once we decide to respond, this generates a prompt
        # to send to the AI, given a message history
        self.prompt_generator = prompt_generator.PromptGenerator(
            discord_settings=self.settings.discord_settings.get_all(),
            oobabooga_settings=self.settings.oobabooga_settings.get_all(),
            persona=self.persona,
            template_store=self.template_store,
        )

        # tracks of the time spent on responding, success rate, etc.
        self.response_stats = response_stats.AggregateResponseStats(
            fn_get_total_tokens=lambda: self.ooba_client.total_response_tokens
        )

        # generates images, if stable diffusion is configured
        # also includes a UI to regenerate images on demand
        self.image_generator = None
        if self.stable_diffusion_client is not None:
            self.image_generator = image_generator.ImageGenerator(
                ooba_client=self.ooba_client,
                persona_settings=self.settings.persona_settings.get_all(),
                prompt_generator=self.prompt_generator,
                sd_settings=self.settings.stable_diffusion_settings.get_all(),
                stable_diffusion_client=self.stable_diffusion_client,
                template_store=self.template_store,
            )

        # if a bot sees itself repeating a message over and over,
        # it will keep doing so forever.  This attempts to fix that.
        # by looking for repeated responses, and deciding how far
        # back in history the bot can see.
        self.repetition_tracker = repetition_tracker.RepetitionTracker(
            repetition_threshold=self.settings.REPETITION_TRACKER_THRESHOLD
        )

        self.bot_commands = bot_commands.BotCommands(
            decide_to_respond=self.decide_to_respond,
            repetition_tracker=self.repetition_tracker,
            persona=self.persona,
            discord_settings=self.settings.discord_settings.get_all(),
            template_store=self.template_store,
        )

        self.discord_bot = None

        if self._our_own_main():
            # if we're running as a worker thread in another process,
            # we can't (and shouldn't) register
            def exit_handler(signum, _frame):
                sig_name = signal.Signals(signum).name
                fancy_logger.get().info("Received signal %s, exiting...", sig_name)
                self.response_stats.write_stat_summary_to_log()
                sys.exit(0)

            signal.signal(signal.SIGINT, exit_handler)
            signal.signal(signal.SIGTERM, exit_handler)

    def _prepare_connections(self):
        ########################################################
        # Test connection to services
        for client in [self.ooba_client, self.stable_diffusion_client]:
            if client is None:
                continue

            fancy_logger.get().info("%s is at %s", client.service_name, client.base_url)
            try:
                client.test_connection()
                fancy_logger.get().info("Connected to %s!", client.service_name)
            except (ValueError, http_client.OobaHttpClientError) as err:
                fancy_logger.get().error(
                    "Could not connect to %s server: [%s]",
                    client.service_name,
                    client.base_url,
                )
                fancy_logger.get().error("Please check the URL and try again.")
                if err.__cause__ is not None:
                    fancy_logger.get().error("Reason: %s", err.__cause__)
                if self._our_own_main():
                    sys.exit(1)
                else:
                    raise

        ########################################################
        # Connect to Discord

        fancy_logger.get().info("Connecting to Discord... ")
        self.discord_bot = discord_bot.DiscordBot(
            bot_commands=self.bot_commands,
            decide_to_respond=self.decide_to_respond,
            discord_settings=self.settings.discord_settings.get_all(),
            ooba_client=self.ooba_client,
            image_generator=self.image_generator,
            persona=self.persona,
            prompt_generator=self.prompt_generator,
            repetition_tracker=self.repetition_tracker,
            response_stats=self.response_stats,
        )

    async def _init_then_start(self):
        """
        Opens HTTP connections to oobabooga and stable diffusion,
        then connects to Discord.  Blocks until the bot is stopped.
        """
        if self.discord_bot is None:
            raise RuntimeError("Discord bot not initialized")

        async with contextlib.AsyncExitStack() as stack:
            for context_manager in [
                self.ooba_client,
                self.stable_diffusion_client,
            ]:
                if context_manager is not None:
                    await stack.enter_async_context(context_manager)

            try:
                await self.discord_bot.start(
                    self.settings.discord_settings.get_str("discord_token")
                )
            except discord.LoginFailure as err:
                fancy_logger.get().error("Could not log in to Discord: %s", err)
                fancy_logger.get().error("Please check the token and try again.")
                if self._our_own_main():
                    sys.exit(1)
            finally:
                await self.discord_bot.close()
        fancy_logger.get().info("Disconnected from Discord.")

    def stop(self, wait_timeout: float = 5.0) -> None:
        """
        Stops the bot, if it's running.  Safe to be called
        from a separate thread from the one that called run().

        Blocks until the bot is gracefully stopped.
        """
        if self.discord_bot is None:
            return

        async def close_and_set() -> None:
            if self.discord_bot is None:
                return
            await self.discord_bot.close()

        with self.startup_lock:
            try:
                # if discord is already stopped, then their .loop is set
                # _MissingSentinel.  So instead check if it's still connected.
                if self.discord_bot.is_closed():
                    fancy_logger.get().info("Discord bot already stopped.")
                    return

                future = asyncio.run_coroutine_threadsafe(
                    close_and_set(),
                    self.discord_bot.loop,
                )
                future.result(timeout=wait_timeout)
            except RuntimeError:
                # this can happen if the main application is shutting down
                fancy_logger.get().warning("Discord bot is already stopped.")
            except concurrent.futures.TimeoutError:
                fancy_logger.get().warning(
                    "Discord bot did not stop in time, it might be busted."
                )

    @classmethod
    def test_discord_token(cls, discord_token: str) -> bool:
        """
        Tests a discord token to see if it's valid.  Can be called from any thread.

        Requires Internet connectivity.

        Returns True if it was able to connect with the token, False if it isn't.
        """
        return asyncio.run(discord_utils.test_discord_token(discord_token))

    @classmethod
    def generate_invite_url(cls, discord_token: str) -> str:
        """
        Generates an invite URL for the bot with the given token.

        Can be called from any thread.  Does not require Internet connectivity.
        """
        ai_user_id = discord_utils.get_user_id_from_token(discord_token)
        return discord_utils.generate_invite_url(ai_user_id)


# pylint: enable=W0201


def main():
    # create the object and load our settings
    oobabot = Oobabot(sys.argv[1:])
    # start the loop
    oobabot.start()



# -*- coding: utf-8 -*-
"""
Logging with colors
"""

import html
import logging
import textwrap
import typing

FOREGROUND_COLORS = {
    "black": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
}

HTML_HEADER = textwrap.dedent(
    """
    <!DOCTYPE html>
    <html lang="en">
    <head>
    <meta charset="utf-8">
    <title>oobabot logs</title>
    <style>
    body {
        background-color: #0C0C0C;
        color: #CCCCCC;
        font-family: Consolas, Lucida Console, monospace;
    }
    .oobabot-red {
        color: #C50F1F;
    }
    .oobabot-yellow {
        color: #C19C00;
    }
    .oobabot-cyan {
        color: #3A96DD;
    }
    .oobabot-white {
        color: #CCCCCC;
    }
    </style>
    </head>
    <body><div class="oobabot-log">
    """
)
HTML_RECORD_SEPARATOR = "\n<br>"
HTML_FOOTER = "</div></body></html>"


def apply_color_console(color: str, text: str) -> str:
    return f"\033[{FOREGROUND_COLORS[color]}m{text}\033[0m"


def apply_color_html(color: str, text: str) -> str:
    return f"<span class='oobabot-{color}'>{text}</span>"


def make_coloring_book(
    fn_apply_color: typing.Callable[[str, str], str]
) -> typing.Dict[int, str]:
    prefix = f'{fn_apply_color("yellow", "%(asctime)s")} %(levelname)s '
    msg = "%(message)s"
    return {
        logging.DEBUG: prefix + fn_apply_color("cyan", msg),
        logging.INFO: prefix + fn_apply_color("white", msg),
        logging.WARNING: prefix + fn_apply_color("yellow", msg),
        logging.ERROR: prefix + fn_apply_color("red", msg),
        logging.CRITICAL: prefix + fn_apply_color("red", msg),
    }


class ColorfulLoggingFormatter(logging.Formatter):
    """
    Logging formatter that adds colors to the log levels.
    """

    def __init__(
        self,
        coloring_book: typing.Dict[int, str],
        fn_format_message: typing.Optional[
            typing.Callable[[typing.Optional[typing.Any]], typing.Optional[typing.Any]]
        ] = None,
    ) -> None:
        super().__init__()
        self.formatters = {}
        for logging_level, fmt_color in coloring_book.items():
            self.formatters[logging_level] = logging.Formatter(fmt_color)
        self.fn_format_message = fn_format_message

    def format(self, record: logging.LogRecord) -> str:
        if self.fn_format_message:
            record = logging.makeLogRecord(record.__dict__)
            record.msg = self.fn_format_message(record.msg)
            # record.args is a tuple.  Call self.fn_format_message for each
            # element of the tuple, and then reassemble the tuple.
            if record.args:
                record.args = tuple(self.fn_format_message(arg) for arg in record.args)

        formatter = self.formatters.get(record.levelno)
        if formatter:
            result = formatter.format(record)

            # From the Python docs for logging.Formatter():
            # ...you should be careful if you have more than one Formatter
            # subclass which customizes the formatting of exception information.
            # In this case, you will have to clear the cached value (by setting
            # the exc_text attribute to None) after a formatter has done its
            # formatting, so that the next formatter to handle the event doesn’t
            # use the cached value, but recalculates it afresh.
            record.exc_text = None
            return result
        return record.getMessage()


def get(name: str = "oobabot") -> logging.Logger:
    return logging.getLogger(name)


def do_escape(msg: typing.Optional[typing.Any]) -> typing.Optional[typing.Any]:
    if msg is None:
        return None
    if not isinstance(msg, str):
        return msg
    result = html.escape(msg)
    return result


def init_logging(
    level: typing.Union[int, str],
    log_to_console: bool = True,
) -> None:
    logger = get()
    logger.setLevel(level)

    if log_to_console:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        console_handler.setFormatter(
            ColorfulLoggingFormatter(
                coloring_book=make_coloring_book(apply_color_console),
            )
        )
        logger.addHandler(console_handler)

    recent_logs.setLevel(level)
    recent_logs.setFormatter(
        ColorfulLoggingFormatter(
            coloring_book=make_coloring_book(apply_color_html),
            fn_format_message=do_escape,
        )
    )
    logger.addHandler(recent_logs)


# the following class was modified from O'Reilly's Python Cookbook,
# chapter 5, section 19.  Its use is allowed under this license:
# Copyright (c) 2001, Sébastien Keim
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#    * Neither the name of the <ORGANIZATION> nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
class RingBuffer:
    """
    A generic ring buffer.
    """

    def __init__(self, size_max: int):
        self.cur = 0
        self.max = size_max
        self.data: typing.List[str] = []

    class _FullRingBuffer:
        """
        Class implementing the RingBuffer when it's full.
        With python class magic, this class is swapped in when the
        buffer becomes full.
        """

        cur: int
        max: int
        data: typing.List[str]

        def append(self, val: str) -> None:
            """
            Append an element overwriting the oldest one.
            """
            self.data[self.cur] = val
            self.cur = (self.cur + 1) % self.max

        def get(self) -> typing.List[str]:
            """
            Return a list of elements from the oldest to the newest.
            """
            return self.data[self.cur :] + self.data[: self.cur]

        def size(self) -> int:
            """
            Return the size of the buffer.
            """
            return self.max

    def append(self, val: str) -> None:
        """
        Append an element at the end of the buffer.
        """
        self.data.append(val)
        if len(self.data) == self.max:
            self.cur = 0
            # Permanently change self's class from non-full to full
            self.__class__ = self._FullRingBuffer

    def get(self) -> typing.List[str]:
        """
        Return a list of elements from the oldest to the newest.
        """
        return self.data

    def size(self) -> int:
        """
        Return the number of elements currently in the buffer.
        """
        return len(self.data)


# end of O'Reilly code


class RingBufferedHandler(logging.Handler):
    """
    A singleton logging handler that stores the last N log messages in a ring buffer.
    """

    def __init__(self, buffer_size: int = 45) -> None:
        super().__init__()
        self.buffer = RingBuffer(buffer_size)

    def emit(self, record: logging.LogRecord) -> None:
        self.buffer.append(self.format(record))

    def get_all(self) -> typing.List[str]:
        return self.buffer.get()


recent_logs = RingBufferedHandler()



# -*- coding: utf-8 -*-
"""
Templates for text generated by the bot.  There are two types:
 - templates used to generate prompts for the AI
 - templates used to generate UI messages for the user
"""

import enum
import functools
import textwrap
import typing


@functools.total_ordering
class Templates(enum.Enum):
    """
    Enumeration of all different templates.
    """

    COMMAND_LOBOTOMIZE_RESPONSE = "command_lobotomize_response"

    IMAGE_DETACH = "image_detach"
    IMAGE_CONFIRMATION = "image_confirmation"
    IMAGE_GENERATION_ERROR = "image_generation_error"
    IMAGE_UNAUTHORIZED = "image_unauthorized"

    # prompts to the AI to generate text responses
    PROMPT = "prompt"
    PROMPT_HISTORY_LINE = "prompt_history_line"
    PROMPT_IMAGE_COMING = "prompt_image_coming"

    # prompts to the AI to generate image keywords
    PROMPT_IMAGE_KEYWORDS = "prompt_image_keywords"

    def __str__(self) -> str:
        return self.value

    def __lt__(self, other: "Templates") -> bool:
        return str(self.value) < str(other.value)


class TemplateToken(str, enum.Enum):
    """
    Enumeration of all tokens used in templates.
    Tokens are variable substitutions into the templates.
    """

    AI_NAME = "AI_NAME"
    PERSONA = "PERSONA"
    IMAGE_COMING = "IMAGE_COMING"
    IMAGE_PROMPT = "IMAGE_PROMPT"
    MESSAGE_HISTORY = "MESSAGE_HISTORY"
    USER_MESSAGE = "USER_MESSAGE"
    USER_NAME = "USER_NAME"

    def __str__(self):
        return "{" + self.value + "}"


class TemplateStore:
    """
    Data object storing all template definitions and default values.
    """

    # mapping of template names to tokens allowed in that template
    #  key: template name
    #  value: tuple of (list of tokens, description, is_an_ai_prompt)
    TEMPLATES: typing.Dict[
        Templates, typing.Tuple[typing.List[TemplateToken], str, bool]
    ] = {
        Templates.COMMAND_LOBOTOMIZE_RESPONSE: (
            [
                TemplateToken.AI_NAME,
                TemplateToken.USER_NAME,
            ],
            "Displayed in Discord after a successful /lobotomize command.  "
            + "Both the discord users and the bot AI will see this message.",
            False,
        ),
        Templates.PROMPT: (
            [
                TemplateToken.AI_NAME,
                TemplateToken.IMAGE_COMING,
                TemplateToken.MESSAGE_HISTORY,
                TemplateToken.PERSONA,
            ],
            "The main prompt sent to Oobabooga to generate a response from "
            + "the bot AI.  The AI's reply to this prompt will be sent to "
            + "discord as the bot's response.",
            True,
        ),
        Templates.PROMPT_HISTORY_LINE: (
            [
                TemplateToken.USER_MESSAGE,
                TemplateToken.USER_NAME,
            ],
            "Part of the AI response-generation prompt, this is used to "
            + "render a single line of chat history.  A list of these, "
            + "one for each past chat message, will become {MESSAGE_HISTORY} "
            + "and inserted into the main prompt",
            True,
        ),
        Templates.PROMPT_IMAGE_COMING: (
            [
                TemplateToken.AI_NAME,
            ],
            "Part of the AI response-generation prompt, this is used to "
            + "inform the AI that it is in the process of generating an "
            + "image.",
            True,
        ),
        Templates.PROMPT_IMAGE_KEYWORDS: (
            [
                TemplateToken.AI_NAME,
                TemplateToken.USER_MESSAGE,
            ],
            """
            Sent to Oobabooga, along with the user's image request, to generate
            image keywords.  The AI's response to this prompt will then be sent
            to Stable Diffusion to generate an image.
            """,
            True,
        ),
        Templates.IMAGE_DETACH: (
            [
                TemplateToken.IMAGE_PROMPT,
                TemplateToken.USER_NAME,
            ],
            "Shown in Discord when the user selects to discard an image "
            + "that Stable Diffusion had generated.",
            False,
        ),
        Templates.IMAGE_CONFIRMATION: (
            [
                TemplateToken.IMAGE_PROMPT,
                TemplateToken.USER_NAME,
            ],
            "Shown in Discord when an image is first generated from "
            + "Stable Diffusion.  This should prompt the user to either "
            + "save or discard the image.",
            False,
        ),
        Templates.IMAGE_GENERATION_ERROR: (
            [
                TemplateToken.IMAGE_PROMPT,
                TemplateToken.USER_NAME,
            ],
            "Shown in Discord when the we could not contact Stable Diffusion "
            + "to generate an image.",
            False,
        ),
        Templates.IMAGE_UNAUTHORIZED: (
            [TemplateToken.USER_NAME],
            "Shown in Discord privately to a user if they try to regenerate "
            "an image that was requested by someone else.",
            False,
        ),
    }

    DEFAULT_TEMPLATES: typing.Dict[Templates, str] = {
        Templates.PROMPT: textwrap.dedent(
            """
            You are in a chat room with multiple participants.
            Below is a transcript of recent messages in the conversation.
            Write the next one to three messages that you would send in this
            conversation, from the point of view of the participant named
            {AI_NAME}.

            {PERSONA}

            All responses you write must be from the point of view of
            {AI_NAME}.
            ### Transcript:
            {MESSAGE_HISTORY}
            {IMAGE_COMING}
            """
        ),
        Templates.PROMPT_HISTORY_LINE: textwrap.dedent(
            """
            {USER_NAME} says:
            {USER_MESSAGE}

            """
        ),
        Templates.PROMPT_IMAGE_COMING: textwrap.dedent(
            """
            {AI_NAME}: is currently generating an image, as requested.
            """
        ),
        Templates.PROMPT_IMAGE_KEYWORDS: textwrap.dedent(
            """
            Below is an instruction that describes a task, paired with an input
            that provides further context. Write a response that appropriately
            completes the request.

            ### Instruction:
            The input text asks for an image to be created.  Please create a list
            of keywords to describe that image.

            ### Input:
            {USER_MESSAGE}

            ### Response:
            """
        ),
        Templates.IMAGE_DETACH: textwrap.dedent(
            """
            {USER_NAME} asked for an image with the prompt:
                '{IMAGE_PROMPT}'
            ...but couldn't find a suitable one.
            """
        ),
        Templates.IMAGE_CONFIRMATION: textwrap.dedent(
            """
            {USER_NAME}, is this what you wanted?
            If no choice is made, this message will 💣 self-destruct 💣 in 3 minutes.
            """
        ),
        Templates.IMAGE_GENERATION_ERROR: textwrap.dedent(
            """
            Something went wrong generating your image.  Sorry about that!
            """
        ),
        Templates.IMAGE_UNAUTHORIZED: textwrap.dedent(
            """
            Sorry, only {USER_NAME} can press the buttons.
            """
        ),
        Templates.COMMAND_LOBOTOMIZE_RESPONSE: textwrap.dedent(
            """
            Ummmm... what were we talking about?
            """
        ),
    }

    def __init__(self, settings: dict):
        self.templates: typing.Dict[Templates, TemplateMessageFormatter] = {}
        for template, (tokens, purpose, is_ai_prompt) in self.TEMPLATES.items():
            template_name = str(template)
            template_fmt = settings[template_name]
            if template_fmt is None:
                raise ValueError(f"Template {template_name} has no default format")
            self.add_template(template, template_fmt, tokens, purpose, is_ai_prompt)

    def add_template(
        self,
        template_name: Templates,
        format_str: str,
        allowed_tokens: typing.List[TemplateToken],
        purpose: str,
        is_ai_prompt: bool,
    ):
        self.templates[template_name] = TemplateMessageFormatter(
            template_name,
            format_str,
            allowed_tokens,
            purpose,
            is_ai_prompt,
        )

    def format(
        self, template_name: Templates, format_args: typing.Dict[TemplateToken, str]
    ) -> str:
        return self.templates[template_name].format(format_args)


class TemplateMessageFormatter:
    """
    Validates that templates are safe to run string.format() on, and
    runs string.format()
    """

    def __init__(
        self,
        template_name: Templates,
        template: str,
        allowed_tokens: typing.List[TemplateToken],
        purpose: str,
        is_ai_prompt: bool,
    ):
        self._validate_format_string(template_name, template, allowed_tokens)
        self.template_name = template_name
        self.template = template
        self.allowed_tokens = allowed_tokens
        self.purpose = purpose
        self.is_ai_prompt = is_ai_prompt

    def __str__(self):
        return self.template

    def format(self, format_args: typing.Dict[TemplateToken, str]) -> str:
        return self.template.format(**format_args)

    @staticmethod
    def _validate_format_string(
        template_name: Templates,
        format_str: str,
        allowed_args: typing.List[TemplateToken],
    ):
        def find_all_ch(string: str, char: str) -> typing.Generator[int, None, None]:
            # find all indices of ch in s
            for i, letter in enumerate(string):
                if letter == char:
                    yield i

        # raises if fmt_string contains any args not in allowed_args
        allowed_close_brace_indices: typing.Set[int] = set()

        for open_brace_idx in find_all_ch(format_str, "{"):
            for allowed_arg in allowed_args:
                idx_end = open_brace_idx + len(allowed_arg) + 1
                next_substr = format_str[open_brace_idx : idx_end + 1]
                if next_substr == "{" + allowed_arg + "}":
                    allowed_close_brace_indices.add(idx_end)
                    break
            else:
                raise ValueError(
                    f"invalid template: {template_name} contains "
                    + f"an argument not in {allowed_args}"
                )
        for close_brace_idx in find_all_ch(format_str, "}"):
            if close_brace_idx not in allowed_close_brace_indices:
                raise ValueError(
                    f"invalid template: {template_name} contains "
                    + f"an argument not in {allowed_args}"
                )



# -*- coding: utf-8 -*-
"""
Documents all the settings for the bot.  Allows for settings
to be loaded from the environment, command line and config file.

Methods:
    - load:
        loads settings from the environment, command line and config file

    - write_to_stream:
        writes the current config file to the given stream

    - write_to_file:
        writes the current config file to the given file path

Attributes:
    - setting_groups:
        a list of all the setting groups

    - discord_settings:
        a setting group for settings related to Discord

    - oobabooga_settings:
        a setting group for settings related to the oobabooga API

    - stable_diffusion_settings:
        a setting group for settings related to the stable diffusion API

    - general_settings:
        a setting group for settings that are not included in the config file
"""
import os
import shutil
import sys
import textwrap
import typing

from oobabot import templates
import oobabot.overengineered_settings_parser as oesp


def _console_wrapped(message):
    width = shutil.get_terminal_size().columns
    return "\n".join(textwrap.wrap(message, width))


def _make_template_comment(
    name: str,
    tokens_desc_tuple: typing.Tuple[typing.List[templates.TemplateToken], str, bool],
) -> typing.List[str]:
    return [
        f"Path to a file containing the {name} template.",
        tokens_desc_tuple[1],
        ".",
        f"Allowed tokens: {', '.join([str(t) for t in tokens_desc_tuple[0]])}",
        ".",
    ]


class Settings:
    """
    User=customizable settings for the bot.  Reads from
    environment variables and command line arguments.
    """

    ############################################################
    # This section is for constants which are not yet
    # customizable by the user.

    # This is a table of the probability that the bot will respond
    # in an unsolicited manner (i.e. it isn't specifically pinged)
    # to a message, based on how long ago it was pinged in that
    # same channel.
    TIME_VS_RESPONSE_CHANCE: typing.List[typing.Tuple[float, float]] = [
        # (seconds, base % chance of an unsolicited response)
        (60.0, 0.90),
        (120.0, 0.70),
        (60.0 * 5, 0.50),
    ]

    # increased chance of responding to a message if it ends with
    # a question mark or exclamation point
    DECIDE_TO_RESPOND_INTERROBANG_BONUS = 0.3

    # number of times in a row that the bot will repeat itself
    # before the repetition tracker will take action
    REPETITION_TRACKER_THRESHOLD = 1

    OOBABOOGA_DEFAULT_REQUEST_PARAMS: oesp.SettingDictType = {
        "max_new_tokens": 250,
        "do_sample": True,
        "temperature": 1.3,
        "top_p": 0.1,
        "typical_p": 1,
        "repetition_penalty": 1.18,
        "top_k": 40,
        "min_length": 0,
        "no_repeat_ngram_size": 0,
        "num_beams": 1,
        "penalty_alpha": 0,
        "length_penalty": 1,
        "early_stopping": False,
        "seed": -1,
        "add_bos_token": True,
        "truncation_length": 730,
        "ban_eos_token": False,
        "skip_special_tokens": True,
        "stopping_strings": [],
    }

    # set default negative prompts to make it more difficult
    # to create content against the discord TOS
    # https://discord.com/guidelines

    # use this prompt for "age_restricted" Discord channels
    #  i.e. channel.nsfw is true
    DEFAULT_SD_NEGATIVE_PROMPT_NSFW: str = "animal harm, suicide, loli"

    # use this prompt for non-age-restricted channels
    DEFAULT_SD_NEGATIVE_PROMPT: str = DEFAULT_SD_NEGATIVE_PROMPT_NSFW + ", nsfw"

    DEFAULT_SD_REQUEST_PARAMS: typing.Dict[str, typing.Union[bool, int, str]] = {
        # default values are commented out
        #
        # "do_not_save_samples": False,
        #    This is a privacy concern for the users of the service.
        #    We don't want to save the generated images anyway, since they
        #    are going to be on Discord.  Also, we don't want to use the
        #    disk space.
        "do_not_save_samples": True,
        #
        # "do_not_save_grid": False,
        #    Sames as above.
        "do_not_save_grid": True,
        "negative_prompt": DEFAULT_SD_NEGATIVE_PROMPT,
        "negative_prompt_nsfw": DEFAULT_SD_NEGATIVE_PROMPT_NSFW,
        "steps": 30,
        "width": 512,
        "height": 512,
        "sampler": "",
    }
    # words to look for in the prompt to indicate that the user
    # wants to generate an image
    DEFAULT_IMAGE_WORDS: typing.List[str] = [
        "draw me",
        "drawing",
        "photo",
        "pic",
        "picture",
        "image",
        "sketch",
    ]

    # ENVIRONMENT VARIABLES ####
    DISCORD_TOKEN_ENV_VAR: str = "DISCORD_TOKEN"
    OOBABOT_PERSONA_ENV_VAR: str = "OOBABOT_PERSONA"

    def __init__(self):
        self._settings = None

        self.arg_parser = None
        self.setting_groups: typing.List[oesp.ConfigSettingGroup] = []

        ###########################################################
        # General Settings
        #  won't be included in the config.yaml

        self.general_settings = oesp.ConfigSettingGroup(
            "General Settings", include_in_yaml=False
        )
        self.setting_groups.append(self.general_settings)

        self.general_settings.add_setting(
            oesp.ConfigSetting[bool](
                name="help",
                default=False,
                description_lines=[],
                cli_args=["-h", "--help"],
            )
        )
        # read a path to a config file from the command line
        self.general_settings.add_setting(
            oesp.ConfigSetting[str](
                name="config",
                default="config.yml",
                description_lines=[
                    "Path to a config file to read settings from.",
                    "Command line settings will override settings in this file.",
                ],
                cli_args=["-c", "--config"],
            )
        )
        self.general_settings.add_setting(
            oesp.ConfigSetting[bool](
                name="generate_config",
                default=False,
                description_lines=[
                    textwrap.dedent(
                        """
                        If set, oobabot will print its configuration as a .yml file,
                        then exit.  Any command-line settings also passed will be
                        reflected in this file.
                        """
                    )
                ],
            )
        )
        self.general_settings.add_setting(
            oesp.ConfigSetting[bool](
                name="invite_url",
                default=False,
                description_lines=[
                    textwrap.dedent(
                        """
                        Print a URL which can be used to invite the
                        bot to a Discord server.  Requires that
                        the Discord token is set.
                        """
                    ),
                ],
                cli_args=["--invite-url"],
            )
        )

        ###########################################################
        # Persona Settings

        self.persona_settings = oesp.ConfigSettingGroup("Persona")
        self.setting_groups.append(self.persona_settings)

        self.persona_settings.add_setting(
            oesp.ConfigSetting[str](
                name="ai_name",
                default="oobabot",
                description_lines=[
                    "Name the AI will use to refer to itself",
                ],
            )
        )
        self.persona_settings.add_setting(
            oesp.ConfigSetting[str](
                name="persona",
                default=os.environ.get(self.OOBABOT_PERSONA_ENV_VAR, ""),
                description_lines=[
                    textwrap.dedent(
                        f"""
                        This prefix will be added in front of every user-supplied
                        request.  This is useful for setting up a 'character' for the
                        bot to play.  Alternatively, this can be set with the
                        {self.OOBABOT_PERSONA_ENV_VAR} environment variable.
                        """
                    )
                ],
                show_default_in_yaml=False,
            )
        )
        # path to a json or txt file containing persona
        self.persona_settings.add_setting(
            oesp.ConfigSetting[str](
                name="persona_file",
                default="",
                description_lines=[
                    textwrap.dedent(
                        """
                        Path to a file containing a persona.  This can be just a
                        single string, a json file in the common "tavern" formats,
                        or a yaml file in the Oobabooga format.

                        With a single string, the persona will be set to that string.

                        Otherwise, the ai_name and persona will be overwritten with
                        the values in the file.  Also, the wakewords will be
                        extended to include the character's own name.
                        """
                    )
                ],
                include_in_argparse=False,
            )
        )
        self.persona_settings.add_setting(
            oesp.ConfigSetting[typing.List[str]](
                name="wakewords",
                default=["oobabot"],
                description_lines=[
                    textwrap.dedent(
                        """
                        One or more words that the bot will listen for.
                        The bot will listen in all discord channels it can
                        access for one of these words to be mentioned, then reply
                        to any messages it sees with a matching word.
                        The bot will always reply to @-mentions and
                        direct messages, even if no wakewords are supplied.
                        """
                    )
                ],
            )
        )

        ###########################################################
        # Discord Settings

        self.discord_settings = oesp.ConfigSettingGroup("Discord")
        self.setting_groups.append(self.discord_settings)

        self.discord_settings.add_setting(
            oesp.ConfigSetting[str](
                name="discord_token",
                default=os.environ.get(self.DISCORD_TOKEN_ENV_VAR, ""),
                description_lines=[
                    textwrap.dedent(
                        f"""
                        Token to log into Discord with.  For security purposes
                        it's strongly recommended that you set this via the
                        {self.DISCORD_TOKEN_ENV_VAR} environment variable
                        instead, if possible.
                        """
                    )
                ],
                cli_args=["--discord-token"],
                show_default_in_yaml=False,
                place_default_in_yaml=True,
            )
        )
        self.discord_settings.add_setting(
            oesp.ConfigSetting[bool](
                name="dont_split_responses",
                default=False,
                description_lines=[
                    textwrap.dedent(
                        """
                        Post the entire response as a single message, rather than
                        splitting it into separate messages by sentence.
                        """
                    )
                ],
            )
        )
        self.discord_settings.add_setting(
            oesp.ConfigSetting[int](
                name="history_lines",
                default=7,
                description_lines=[
                    textwrap.dedent(
                        """
                        Number of lines of chat history the AI will see
                        when generating a response.
                        """
                    )
                ],
            )
        )
        self.discord_settings.add_setting(
            oesp.ConfigSetting[bool](
                name="ignore_dms",
                default=False,
                description_lines=[
                    textwrap.dedent(
                        """
                        If set, the bot will not respond to direct messages.
                        """
                    )
                ],
            )
        )
        # log level
        self.discord_settings.add_setting(
            oesp.ConfigSetting[str](
                name="log_level",
                default="DEBUG",
                description_lines=[
                    "Set the log level.  Valid values are: ",
                    "CRITICAL, ERROR, WARNING, INFO, DEBUG",
                ],
                include_in_argparse=False,
            )
        )
        self.discord_settings.add_setting(
            oesp.ConfigSetting[bool](
                name="reply_in_thread",
                default=False,
                description_lines=[
                    textwrap.dedent(
                        """
                        If set, the bot will generate a thread to respond in
                        if it is not already in one.
                        """
                    )
                ],
            )
        )
        self.discord_settings.add_setting(
            oesp.ConfigSetting[typing.List[str]](
                name="stop_markers",
                default=[
                    "### End of Transcript ###<|endoftext|>",
                    "<|endoftext|>",
                ],
                description_lines=[
                    textwrap.dedent(
                        """
                        A list of strings that will cause the bot to stop
                        generating a response when encountered.
                        """
                    )
                ],
                include_in_argparse=False,
            )
        )
        self.discord_settings.add_setting(
            oesp.ConfigSetting[bool](
                name="stream_responses",
                default=False,
                description_lines=[
                    textwrap.dedent(
                        """
                        FEATURE PREVIEW: Stream responses into a single message
                        as they are generated.
                        Note: may be janky
                        """
                    )
                ],
            )
        )

        ###########################################################
        # Oobabooga Settings

        self.oobabooga_settings = oesp.ConfigSettingGroup("Oobabooga")
        self.setting_groups.append(self.oobabooga_settings)

        self.oobabooga_settings.add_setting(
            oesp.ConfigSetting[str](
                name="base_url",
                default="ws://localhost:5005",
                description_lines=[
                    textwrap.dedent(
                        """
                        Base URL for the oobabooga instance.  This should be
                        ws://hostname[:port] for plain websocket connections,
                        or wss://hostname[:port] for websocket connections over TLS.
                        """
                    )
                ],
            )
        )
        self.oobabooga_settings.add_setting(
            oesp.ConfigSetting[bool](
                name="log_all_the_things",
                default=False,
                description_lines=[
                    textwrap.dedent(
                        """
                        Print all AI input and output to STDOUT.
                        """
                    )
                ],
            )
        )
        # get a regex for filtering a message
        self.oobabooga_settings.add_setting(
            oesp.ConfigSetting[str](
                name="message_regex",
                default="",
                description_lines=[
                    textwrap.dedent(
                        """
                        A regex that will be used to extract message lines
                        from the AI's output.  The first capture group will
                        be used as the message.  If this is not set, the
                        entire output will be used as the message.
                        """
                    )
                ],
            )
        )
        self.oobabooga_settings.add_setting(
            oesp.ConfigSetting[oesp.SettingDictType](
                name="request_params",
                default=self.OOBABOOGA_DEFAULT_REQUEST_PARAMS,
                description_lines=[
                    textwrap.dedent(
                        """
                        A dictionary which will be passed straight through to
                        Oobabooga on every request.  Feel free to add additional
                        simple parameters here as Oobabooga's API evolves.
                        See Oobabooga's documentation for what these parameters
                        mean.
                        """
                    )
                ],
                include_in_argparse=False,
                show_default_in_yaml=False,
                place_default_in_yaml=True,
            )
        )

        ###########################################################
        # Stable Diffusion Settings

        self.stable_diffusion_settings = oesp.ConfigSettingGroup("Stable Diffusion")
        self.setting_groups.append(self.stable_diffusion_settings)

        self.stable_diffusion_settings.add_setting(
            oesp.ConfigSetting[typing.List[str]](
                name="image_words",
                default=self.DEFAULT_IMAGE_WORDS,
                description_lines=[
                    textwrap.dedent(
                        """
                        When one of these words is used in a message, the bot will
                        generate an image.
                        """
                    )
                ],
            )
        )
        self.stable_diffusion_settings.add_setting(
            oesp.ConfigSetting[str](
                name="stable_diffusion_url",
                default="",
                description_lines=[
                    textwrap.dedent(
                        """
                        URL for an AUTOMATIC1111 Stable Diffusion server.
                        """
                    )
                ],
            )
        )
        self.stable_diffusion_settings.add_setting(
            oesp.ConfigSetting[str](
                name="extra_prompt_text",
                default="",
                description_lines=[
                    textwrap.dedent(
                        """
                        This will be appended to every image generation prompt
                        sent to Stable Diffusion.
                        """
                    )
                ],
            )
        )
        self.stable_diffusion_settings.add_setting(
            oesp.ConfigSetting[oesp.SettingDictType](
                name="request_params",
                default=self.DEFAULT_SD_REQUEST_PARAMS,
                description_lines=[
                    textwrap.dedent(
                        """
                        A dictionary which will be passed straight through to
                        Stable Diffusion on every request.  Feel free to add additional
                        simple parameters here as Stable Diffusion's API evolves.
                        See Stable Diffusion's documentation for what these parameters
                        mean.
                        """
                    )
                ],
                include_in_argparse=False,
                show_default_in_yaml=False,
                place_default_in_yaml=True,
            )
        )
        self.stable_diffusion_settings.add_setting(
            oesp.ConfigSetting[bool](
                name="use_ai_generated_keywords",
                default=False,
                description_lines=[
                    textwrap.dedent(
                        """
                        FEATURE PREVIEW: If set, the bot will ask Oobabooga to generate
                        image keywords from a user's message.  It will then pass the
                        keywords that Oobabooga produces to Stable Diffusion to finally
                        generate an image.
                        Otherwise, the bot will simply extract keywords directly
                        from the user's message using a simple regex.
                        """
                    )
                ],
            )
        )

        ###########################################################
        # Template Settings

        self.template_settings = oesp.ConfigSettingGroup(
            "Template",
            description="UI and AI request templates",
            include_in_argparse=False,
        )
        self.setting_groups.append(self.template_settings)

        for template, tokens_desc_tuple in templates.TemplateStore.TEMPLATES.items():
            _, _, is_ai_prompt = tokens_desc_tuple

            self.template_settings.add_setting(
                oesp.ConfigSetting[str](
                    name=str(template),
                    default=templates.TemplateStore.DEFAULT_TEMPLATES[template],
                    description_lines=_make_template_comment(
                        str(template), tokens_desc_tuple
                    ),
                    include_in_yaml=is_ai_prompt,
                )
            )

        self._add_deprecated_settings()

    def _add_deprecated_settings(self) -> None:
        ###########################################################
        # Deprecated Settings
        # These used to be part of the Stable Diffusion section,
        # but now are covered in the stable_diffusion->request_params
        # section of the config.yml file.
        #
        # These are still here for backwards compatibility, but
        # will be removed in a future release.
        #
        # They're being moved since they're redundant with the
        # request_params values, and it's confusing to have those
        # set in two separate places.
        #
        # The settings won't be written to the config.yaml template,
        # as they're already covered in the request_params section.

        def set_sd_parm(param: str, value: oesp.SettingValueType):
            if not value:
                return
            setting = self.stable_diffusion_settings.get_setting("request_params")

            # get returns a copy of the settings dict, so we need to
            # push it back after we make a change
            setting_dict = setting.get()
            setting_dict[param] = value
            setting.set_value(setting_dict)

        self.deprecated_settings = oesp.ConfigSettingGroup(
            "Deprecated Settings",
            include_in_yaml=False,
            description="These settings are deprecated and will be removed in "
            + "a future release.  Please set them with config.yml instead.",
        )
        self.setting_groups.append(self.deprecated_settings)

        self.deprecated_settings.add_setting(
            oesp.ConfigSetting[int](
                name="diffusion_steps",
                default=0,
                description_lines=[
                    textwrap.dedent(
                        """
                        Number of diffusion steps to take when generating an image.
                        """
                    )
                ],
                fn_on_set=lambda x: set_sd_parm("steps", int(x)),
            )
        )
        self.deprecated_settings.add_setting(
            oesp.ConfigSetting[int](
                name="image_height",
                default=0,
                description_lines=[
                    textwrap.dedent(
                        """
                        Size of images to generate.  This is the height of the image
                        in pixels.
                        """
                    )
                ],
                fn_on_set=lambda x: set_sd_parm("height", int(x)),
            )
        )
        self.deprecated_settings.add_setting(
            oesp.ConfigSetting[int](
                name="image_width",
                default=0,
                description_lines=[
                    textwrap.dedent(
                        """
                        Size of images to generate.  This is the width of the image
                        in pixels.
                        """
                    )
                ],
                fn_on_set=lambda x: set_sd_parm("width", int(x)),
            )
        )
        self.deprecated_settings.add_setting(
            oesp.ConfigSetting[str](
                name="stable_diffusion_sampler",
                default="",
                description_lines=[
                    textwrap.dedent(
                        """
                        Sampler to use when generating images.  If not specified,
                        the one set on the AUTOMATIC1111 server will be used.
                        """
                    )
                ],
                cli_args=["--stable-diffusion-sampler", "--sd-sampler"],
                fn_on_set=lambda x: set_sd_parm("sampler", str(x)),
            )
        )
        self.deprecated_settings.add_setting(
            oesp.ConfigSetting[str](
                name="sd_negative_prompt",
                default="",
                description_lines=[
                    textwrap.dedent(
                        """
                        Negative prompt to use when generating images.  This will
                        discourage Stable Diffusion from generating images with the
                        specified content.  By default, this is set to follow
                        Discord's TOS.
                        """
                    )
                ],
                fn_on_set=lambda x: set_sd_parm("negative_prompt", str(x)),
            )
        )
        self.deprecated_settings.add_setting(
            oesp.ConfigSetting[str](
                name="sd_negative_prompt_nsfw",
                default="",
                description_lines=[
                    textwrap.dedent(
                        """
                        Negative prompt to use when generating images in a channel
                        marked as 'Age-Restricted'.
                        """
                    )
                ],
                fn_on_set=lambda x: set_sd_parm("negative_prompt_nsfw", str(x)),
            )
        )

    META_INSTRUCTION = (
        "\n\n"
        + "# " * 30
        + textwrap.dedent(
            """
            # Please save this output into ./config.yml
            # edit it to your liking, then run the bot again.
            #
            #  e.g. oobabot --generate-config > config.yml
            #       oobabot
            """
        )
    )

    def write_to_stream(self, out_stream) -> None:
        oesp.write_to_stream(self.setting_groups, out_stream)
        if sys.stdout.isatty():
            print(self.META_INSTRUCTION, file=sys.stderr)
        else:
            print("# oobabot: config.yml output successfully", file=sys.stderr)

    def write_to_file(self, filename: str) -> None:
        oesp.write_to_file(self.setting_groups, filename)

    def _filename_from_args(self, args: typing.List[str]) -> str:
        """
        Get the configuration filename from the command line arguments.
        If none is supplied, return the default.
        """

        # we need to hack this in here because we want to know the filename
        # before we parse the args, so that we can load the config file
        # first and then have the arguments overwrite the config file.
        config_setting = self.general_settings.get_setting("config")
        if args is not None:
            for config_flag in config_setting.cli_args:
                # find the element after config_flag in args
                try:
                    return args[args.index(config_flag) + 1]
                except (ValueError, IndexError):
                    continue
        return config_setting.default

    def load(
        self,
        cli_args: typing.List[str],
        config_file: typing.Optional[str] = None,
    ) -> None:
        """
        Load the config from the command line arguments and config file.

        params:
            cli_args: list of command line arguments to parse
            config_file: path to the config file to load

        cli_args is intended to be used when running from a standalone
        application, while config_file is intended to be used when
        running from inside another process.
        """

        if config_file is None:
            config_file = self._filename_from_args(cli_args)

        self.arg_parser = oesp.load(
            cli_args=cli_args,
            setting_groups=self.setting_groups,
            config_file=config_file,
        )

    def print_help(self):
        """
        Prints CLI usage information to STDOUT.
        """
        if self.arg_parser is None:
            raise ValueError("display_help called before load")

        help_str = self.arg_parser.format_help()
        print(help_str)

        print(
            "\n"
            + _console_wrapped(
                (
                    "Additional settings can be set in config.yml.  "
                    "Use the --generate-config option to print a new "
                    "copy of this file to STDOUT."
                )
            )
        )

        if "" == self.discord_settings.get("discord_token"):
            print(
                "\n"
                + _console_wrapped(
                    (
                        f"Please set the '{self.DISCORD_TOKEN_ENV_VAR}' "
                        "environment variable to your bot's discord token."
                    )
                )
            )
